{
  This demo shows how to convert RTF, RVF, or DocX files to PDF.

  This demo uses LLPDFLib (free open source PDF library)

  TRichView: https://www.trichview.com/
  LLPDFLib: https://github.com/sybrexsys/llPDFLib

  This demo:
  - supports documents with headers, footers, footnotes and endnotes,
  - reads page layout from files
  - supports hyperlinks, both to external URLs and to places in the same document
  - can build table of contents

  The form defined in this unit is inherited from the form defined in
  CustomMakePDFFrm.pas. The parent form implements reading files,
  assigning RVPrint1 parameters, working with the print preview (zooming and
  browsing pages).
  This form implements PDF export.

  The basic version of this demo is in TRichView demos, in the folder
  ThirdParty\Export\LLPDFLib\
  This version of the demo adds a table of contents.

}

unit MainFrm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics,
  Controls, Forms, Dialogs, ExtCtrls, ComCtrls, Buttons, StdCtrls,
  PtblRV, RVScroll, CRVPP, RVPP, RVDocParams, RVStyle, RVFuncs, CRVData,
  RVTypes, RVTable, RVItem,
  RVWithHeaderFooter,
  CustomMakePDFFrm, llPDFTypes, llPDFCanvas,
  llPDFDocument, llPDFOutline, llPDFAction;

type

  TCheckpointLocation = class (TCollectionItem)
    public
      Name: TRVUnicodeString;
      PageIndex, Top: Integer;
  end;
  TLinkToCheckpoint = class (TCollectionItem)
    public
      CheckpointName: TRVUnicodeString;
      PageIndex: Integer;
      Area: TRect;
  end;

  TPDFOutlineNodesArray = array[1..6] of TPDFOutlineNode;

  TfrmMain = class(TfrmCustomMakePDF)
    pdf: TPDFDocument;
    cbTOC: TCheckBox;
    procedure btnExportClick(Sender: TObject);
    procedure RVPrint1DrawCheckpoint(Sender: TCustomRVPrint;
      RVData: TCustomRVData; ItemNo, X, Y: Integer);
    procedure RVPrint1DrawHyperlink(Sender: TCustomRVPrint;
      RVData: TCustomRVData; ItemNo: Integer; R: TRect);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
    function FindCheckpoint(const CheckpointName: TRVUnicodeString;
      var PageIndex, Top: Integer): Boolean;
    procedure SaveLinksToCheckpoints;
    procedure SetPDFURL(RVData: TCustomRVData; ItemNo: Integer; R: TRect);
    procedure DoBuildOutlines(RVData: TCustomRVData;
      var OutlineArray: TPDFOutlineNodesArray);
    procedure BuildOutlines;
    procedure AddHeadingCheckpoints(RVData: TCustomRVData;
      var HeadingIndex: Integer);
  public
    { Public declarations }
    FCurPage: TPDFPage;
    FCurPageIndex: Integer;
    FCheckpoints: TCollection;
    FLinksToCheckpoints: TCollection;
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.dfm}

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  inherited;
  // We will create PDF in 96 DPI. You can change it, if you want.
  PDFPixelsPerInch := 96;
end;

procedure TfrmMain.btnExportClick(Sender: TObject);
var
  i: Integer;
  DocParameters: TRVDocParameters;
  PageSize: TSize;
  ARVStyle: TRVStyle;
  FirstHeadingCheckpointIndex: Integer;
begin
  if not SaveDialog1.Execute then
    exit;

  // Adding checkpoints for all heading (for building PDF outlines)
  FirstHeadingCheckpointIndex := 0;

  if cbTOC.Checked then
    AddHeadingCheckpoints(FDocument.MainDoc.RichView.RVData, FirstHeadingCheckpointIndex);

  // Calculating page size in pixels, with DPI used to measure sizes in document
  DocParameters := FDocument.MainDoc.RichView.DocParameters;
  ARVStyle      := FDocument.MainDoc.RichView.Style;
  PageSize.cx := ARVStyle.RVUnitsToStandardPixels(DocParameters.PageWidth,
    DocParameters.Units);
  PageSize.cy := ARVStyle.RVUnitsToStandardPixels(DocParameters.PageHeight,
    DocParameters.Units);

  pdf.Resolution := PDFPixelsPerInch;
  pdf.FileName := SaveDialog1.FileName;
  pdf.BeginDoc;
  // PDF page size is measured in pixels, with DPI = PDFPixelsPerInch
  pdf.Page[0].Width :=
    MulDiv(PageSize.cx, PDFPixelsPerInch, GetUnitsPixelsPerInch);
  pdf.Page[0].Height :=
    MulDiv(PageSize.cy, PDFPixelsPerInch, GetUnitsPixelsPerInch);

  Screen.Cursor := crHourGlass;

  try
    RVPP.RVPrint := nil; // to avoid repainting and calling events while generating
    FCheckpoints := TCollection.Create(TCheckpointLocation);
    FLinksToCheckpoints := TCollection.Create(TLinkToCheckpoint);
    // Drawing PDF pages
    for i := 0 to RVPrint1.PagesCount - 1 do begin
      ShowText(Format('Collecting PDF pages (%d of %d)... ',
        [i + 1, RVPrint1.PagesCount]));
      if i > 0 then
      begin
        pdf.NewPage;
        pdf.Page[i].Width := pdf.Page[0].Width;
        pdf.Page[i].Height := pdf.Page[0].Height;
      end;
      FCurPage := pdf.Page[i];
      FCurPageIndex := i;
      pdf.Canvas.Font.PixelsPerInch := PDFPixelsPerInch;
      if PDFPixelsPerInch <> Screen.PixelsPerInch then
        {
         This is a workaround:
         LLPDFLib calculates page size, link and checkpoint
         positions using DPI = pdf.Resolution, but it handles drawing to
         pdf.Canvas as if it is in DPI = Screen.PixelsPerInch.
         So we scale the output to DPI = Screen.PixelsPerInch.
         If this problem will be fixed in future versions of LLPDFLib,
         you will be able using DrawPage instead of DrawPreview.
         I tested with LLPDFLib 6.4.0.1389
        }
        RVPrint1.DrawPreview(i + 1, pdf.Canvas, Bounds(0, 0,
          MulDiv(PageSize.cx, Screen.PixelsPerInch, GetUnitsPixelsPerInch),
          MulDiv(PageSize.cy, Screen.PixelsPerInch, GetUnitsPixelsPerInch)))
      else
        RVPrint1.DrawPage(i + 1, pdf.Canvas, True);
    end;
    if cbTOC.Checked then
    begin
      ShowText('Making TOC...');
      BuildOutlines;
    end;
    // Finalizing
    ShowText('Writing PDF file...');
    SaveLinksToCheckpoints;
    pdf.EndDoc;
  finally
    FreeAndNil(FCheckpoints);
    FreeAndNil(FLinksToCheckpoints);
    RVPP.RVPrint := RVPrint1;
    Screen.Cursor := crDefault;
    ShowText('');
  end;
end;
{------------------------------------------------------------------------------}
{ Saving the hyperlink in the RVData's ItemNo-th item in PDF at the location
  specified in R. The location is relative to DocRect.
  If this is a local link (starts from '#'), we add it in FLinksToCheckpoints
  collection. Otherwise, we save the link to PDF as a link to URL. }
procedure TfrmMain.SetPDFURL(RVData: TCustomRVData; ItemNo: Integer; R: TRect);
var Target: String;
begin
  try
    Target := RVData.GetItemTag(ItemNo);
    if Target <> '' then
    begin
      if Target[1] <> '#' then
        FCurPage.SetUrl(R, TRVAnsiString(Target))
      else
        with FLinksToCheckpoints.Add as TLinkToCheckpoint do
        begin
          Area := R;
          PageIndex := FCurPageIndex;
          CheckpointName := Copy(Target, 2, Length(Target)-1);
        end
    end;
  except
  end;
end;
{------------------------------------------------------------------------------}
{ Event: Occurs on drawing a hyperlink. }
procedure TfrmMain.RVPrint1DrawHyperlink(Sender: TCustomRVPrint;
  RVData: TCustomRVData; ItemNo: Integer; R: TRect);
begin
  if FCheckpoints = nil then
    exit;
  SetPDFURL(RVData, ItemNo, R);
end;
{------------------------------------------------------------------------------}
{ Event: Occurs when drawing a checkpoint.
  We add the location of this checkpoint to FCheckpoints collection }
procedure TfrmMain.RVPrint1DrawCheckpoint(Sender: TCustomRVPrint;
  RVData: TCustomRVData; ItemNo, X, Y: Integer);
var CheckpointData: TCheckpointData;
    CPTag: TRVTag;
    CPName: TRVUnicodeString;
    RE: Boolean;
begin
  if FCheckpoints = nil then
    exit;
  CheckpointData := RVData.GetItemCheckpoint(ItemNo);
  RVData.GetCheckpointInfo(CheckpointData, CPTag, CPName, RE);
  with FCheckpoints.Add as TCheckpointLocation do
  begin
    Name := CPName;
    PageIndex := FCurPageIndex;
    Top := Y;
  end;
end;
{------------------------------------------------------------------------------}
{ Returns position of checkpoint with the given name }
function TfrmMain.FindCheckpoint(const CheckpointName: TRVUnicodeString;
  var PageIndex, Top: Integer): Boolean;
var
  i: Integer;
  Item: TCheckpointLocation;
begin
  for i := 0 to FCheckpoints.Count-1 do
  begin
    Item := TCheckpointLocation(FCheckpoints.Items[i]);
    if Item.Name=CheckpointName then
    begin
      PageIndex := Item.PageIndex;
      Top := Item.Top;
      Result := True;
      exit;
    end;
  end;
  Result := False;
end;
{------------------------------------------------------------------------------}
{ Saves all links to checkpoints to PDF }
procedure TfrmMain.SaveLinksToCheckpoints;
var
  i, PageIndex, Top: Integer;
  Item: TLinkToCheckpoint;
begin
  for i := 0 to FLinksToCheckpoints.Count-1 do
  begin
    Item := TLinkToCheckpoint(FLinksToCheckpoints.Items[i]);
    if FindCheckpoint(Item.CheckpointName, PageIndex, Top) then
    begin
      pdf.CurrentPageIndex := Item.PageIndex;
      pdf.Page[Item.PageIndex].SetLinkToPage(Item.Area, PageIndex, Top);
    end;
  end;
end;
{------------------------------------------------------------------------------}
// Functions for building TOC (table of contents) in PDF.
// TOC is built basing on headings
{------------------------------------------------------------------------------}
// Do the specified items have the same heading level?
function ItemsHaveEqualOutlineLevels(RVData: TCustomRVData;
  ItemNo1, ItemNo2: Integer): Boolean;
begin
  Result :=
    RVData.GetRVStyle.ParaStyles[RVData.GetItemPara(ItemNo1)].OutlineLevel =
    RVData.GetRVStyle.ParaStyles[RVData.GetItemPara(ItemNo2)].OutlineLevel;
end;
{------------------------------------------------------------------------------}
// Returns text of a heading in RVData that starts from the ItemNo-th item.
// We assume that this heading can consists of multiple paragraphs.
// We add ': ' in places of paragraph or hard line breaks.
function GetHeadingText(RVData: TCustomRVData; ItemNo: Integer): TRVUnicodeString;
const
  LineBreakHeadingSeparator = ': ';
var
  i: Integer;
begin
  Result := '';
  for i := ItemNo to RVData.ItemCount - 1 do
  begin
    if i > ItemNo then
    begin
      if RVData.IsParaStart(i) and
        not ItemsHaveEqualOutlineLevels(RVData, ItemNo, i) then
        break;
      if RVData.IsFromNewLine(i) then
        Result := Result + LineBreakHeadingSeparator;
    end;
    if RVData.GetItemStyle(i) >= 0 then
      Result := Result + RVData.GetItemTextW(i)
    else if RVData.GetItemStyle(i) = rvsTab then
      Result := Result + ' ';
  end;
end;
{------------------------------------------------------------------------------}
// Adding a checkpoint for each heading
// Note: several paragraphs of the same heading levels are counted as a single
// header
procedure TfrmMain.AddHeadingCheckpoints(RVData: TCustomRVData;
  var HeadingIndex: Integer);
const
  HeadingCheckpointPrefix = '@rvheading@';
var
  i, r, c: Integer;
  Table: TRVTableItemInfo;
begin
  for i := 0 to RVData.ItemCount - 1 do
    if RVData.GetItem(i) is TRVTableItemInfo then
    begin
      Table := TRVTableItemInfo(RVData.GetItem(i));
      for r := 0 to Table.RowCount - 1 do
        for c := 0 to Table.ColCount - 1 do
          if Table.Cells[r, c] <> nil then
             AddHeadingCheckpoints(Table.Cells[r,c].GetRVData, HeadingIndex)
    end
    else if RVData.IsParaStart(i) and (RVData.GetItemCheckpoint(i) = nil) and
      RVData.GetRVStyle.ParaStyles[RVData.GetItemPara(i)].IsHeading(6) and
      ((i = 0) or not ItemsHaveEqualOutlineLevels(RVData, i, i-1)) then
    begin
      RVData.SetCheckpointInfo(i, '', HeadingCheckpointPrefix + IntToStr(HeadingIndex), False);
      inc(HeadingIndex);
    end;
end;
{------------------------------------------------------------------------------}
// Returns an index of the parent heading for the new heading of the specified Level
function FindParentOutline(const OutlineArray: TPDFOutlineNodesArray;
  Level: Integer): Integer;
var
  i: Integer;
begin
  for i := Level - 1 downto 1 do
    if OutlineArray[i] <> nil then
    begin
      Result := i;
      exit;
    end;
  Result := -1;
end;
{------------------------------------------------------------------------------}
// Adds heading to the array
procedure SetOutline(var OutlineArray: TPDFOutlineNodesArray;
  Node: TPDFOutlineNode; Level: Integer);
var
  i: Integer;
begin
  OutlineArray[Level] := Node;
  for i := Level + 1 to 6 do
    OutlineArray[i] := nil;
end;
{------------------------------------------------------------------------------}
// Builds PDF TOC starting from RVData. Called from BuildOutlines.
procedure TfrmMain.DoBuildOutlines(RVData: TCustomRVData;
  var OutlineArray: TPDFOutlineNodesArray);
var
  i, r, c, Level: Integer;
  Table: TRVTableItemInfo;
  CPTag: TRVTag;
  CPName: TRVUnicodeString;
  CPRE: Boolean;
  CPPageIndex, CPTop: Integer;
  HeadingText: TRVUnicodeString;
  Action: TPDFGoToPageAction;
  ParentOutlineIndex: Integer;
begin
  for i := 0 to RVData.ItemCount - 1 do
    if RVData.GetItem(i) is TRVTableItemInfo then
    begin
      Table := TRVTableItemInfo(RVData.GetItem(i));
      for r := 0 to Table.RowCount - 1 do
        for c := 0 to Table.ColCount - 1 do
          if Table.Cells[r, c] <> nil then
             DoBuildOutlines(Table.Cells[r,c].GetRVData, OutlineArray);
    end
    else if RVData.IsParaStart(i) and (RVData.GetItemCheckpoint(i) <> nil) and
      RVData.GetRVStyle.ParaStyles[RVData.GetItemPara(i)].IsHeading(6) and
      ((i = 0) or not ItemsHaveEqualOutlineLevels(RVData, i, i-1)) then
    begin
      Level := RVData.GetRVStyle.ParaStyles[RVData.GetItemPara(i)].OutlineLevel;
      RVData.GetCheckpointInfo(RVData.GetItemCheckpoint(i), CPTag, CPName, CPRE);
      if not FindCheckpoint(CPName, CPPageIndex, CPTop) then
        break;
      Action := TPDFGoToPageAction.Create(pdf.Actions, CPPageIndex, CPTop, True);
      HeadingText := GetHeadingText(RVData, i);
      ParentOutlineIndex := FindParentOutline(OutlineArray, Level);
      if ParentOutlineIndex < 0 then
        SetOutline(OutlineArray,
          pdf.Outlines.Add(nil, HeadingText, Action), Level)
      else
        SetOutline(OutlineArray,
          pdf.Outlines.AddChild(OutlineArray[ParentOutlineIndex],
            HeadingText, Action), Level);
    end;
end;
{------------------------------------------------------------------------------}
// Builds PDF TOC.
procedure TfrmMain.BuildOutlines;
var
  OutlineArray: TPDFOutlineNodesArray;
  i: Integer;
begin
  for i := 1 to 6 do
    OutlineArray[i] := nil;
  DoBuildOutlines(FDocument.MainDoc.RichView.RVData, OutlineArray);
end;

end.

