{
  This form is used in several demo projects which export RTF and RVF files
  to different formats, specifically PDF

  This form implements:
  - loading RTF and RVF files
  - storing them, including all headers and footers, in FDocument
  - displaying print preview of output

  This form is not used directly, it is used as a parent for other forms
}

unit CustomMakePDFFrm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics,
  Controls, Forms, Dialogs, StdCtrls, Buttons,
  ExtCtrls, ComCtrls,
  RVScroll, CRVPP, RVPP, RichView, RVStyle,
  RVWithHeaderFooter, PtblRV, RVFuncs, RVDocParams, CRVData, CRVFData, RVItem, RVTypes, RVTable;

type
  TfrmCustomMakePDF = class(TForm)
    StatusBar1: TStatusBar;
    Panel1: TPanel;
    RVPP: TRVPrintPreview;
    Panel2: TPanel;
    btnOpen: TButton;
    btnExport: TButton;
    cmbZoom: TComboBox;
    btnFirst: TSpeedButton;
    btnPrior: TSpeedButton;
    btnNext: TSpeedButton;
    btnLast: TSpeedButton;
    txtPage: TEdit;
    lblPageCount: TLabel;
    RVPrint1: TRVPrint;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    procedure RVPPZoomChanged(Sender: TObject);
    procedure btnPageNavigateClick(Sender: TObject);
    procedure txtPageKeyPress(Sender: TObject; var Key: Char);
    procedure txtPageExit(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure cmbZoomExit(Sender: TObject);
    procedure cmbZoomKeyPress(Sender: TObject; var Key: Char);
    procedure cmbZoomClick(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
  private
    procedure UpdateZoom;
    procedure UpdatePageNavigator;
    procedure GoToPage;
    procedure AssignVPProperties;
    procedure DoProgress(Sender: TCustomRichView; Operation: TRVLongOperation;
      Stage: TRVProgressStage; PercentDone: Byte);
  protected
    FDocument: TRVReportHelperWithHeaderFooters;
    PDFPixelsPerInch: Integer;
    function GetUnitsPixelsPerInch: Integer;
    procedure LoadFile(const FileName: String; FilterIndex: Integer);
    function ToPDFDPI(Value: Integer): Integer;
    procedure ShowError(const s: String);
    procedure ShowText(const s: String);
  public
    procedure DoItemAction(Sender: TCustomRichView;
      ItemAction: TRVItemAction; Item: TCustomRVItemInfo;
      var Text: TRVUnicodeString; RVData: TCustomRVData);
  end;

implementation

{$R *.dfm}

procedure TfrmCustomMakePDF.FormCreate(Sender: TObject);
begin
  // This value will be used as the output DPI.
  // It will be a DPI of RVPrint1.VirtualPrinter and PDF engine
  PDFPixelsPerInch := 96;
  // Creating storage for the document, including headers and footers
  FDocument := TRVReportHelperWithHeaderFooters.Create;
  FDocument.MainDoc.RichView.OnProgress := DoProgress;
  FDocument.UseStyleTemplates := True;
  // Initializing "virtual printer"
  FDocument.AssignToRVPrint(RVPrint1);
  FDocument.MainDoc.RichView.OnItemAction := DoItemAction;
  RVPrint1.VirtualPrinter.Active := True;
  RVPrint1.FormatPages(rvdoAll);
  // Preparing preview
  RVPP.RVPrint := RVPrint1;
  UpdatePageNavigator;
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.FormDestroy(Sender: TObject);
begin
  FDocument.Free;
end;
{------------------------------------------------------------------------------}
{ Returns DPI used to measure sizes in the document }
function TfrmCustomMakePDF.GetUnitsPixelsPerInch: Integer;
begin
  Result := FDocument.MainDoc.RichView.Style.UnitsPixelsPerInch;
end;
{------------------------------------------------------------------------------}
{ Converting from input to output DPI }
function TfrmCustomMakePDF.ToPDFDPI(Value: Integer): Integer;
begin
  Result := Round(Value * PDFPixelsPerInch / GetUnitsPixelsPerInch);
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.AssignVPProperties;
var
  DocParameters: TRVDocParameters;
  ARVStyle: TRVStyle;
  LPageSize: TSize;
begin
  // Calculating page size in pixels, with DPI = GetUnitsPixelsPerInch
  DocParameters := FDocument.MainDoc.RichView.DocParameters;
  ARVStyle      := FDocument.MainDoc.RichView.Style;
  LPageSize.cx := ARVStyle.RVUnitsToStandardPixels(DocParameters.PageWidth,
    DocParameters.Units);
  LPageSize.cy := ARVStyle.RVUnitsToStandardPixels(DocParameters.PageHeight,
    DocParameters.Units);
  // Assigning RVPrint properties
  RVPrint1.VirtualPrinter.PixelsPerInch := PDFPixelsPerInch;
  RVPrint1.VirtualPrinter.PageWidth := ToPDFDPI(LPageSize.cx);
  RVPrint1.VirtualPrinter.PageHeight := ToPDFDPI(LPageSize.cy);
end;
{========================= Displaying messages ================================}
procedure TfrmCustomMakePDF.ShowError(const s: String);
begin
  Application.MessageBox(PChar(s), 'Error', MB_OK or MB_ICONSTOP);
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.ShowText(const s: String);
begin
  StatusBar1.SimpleText := s;
end;
{======================== File loading ========================================}
procedure TfrmCustomMakePDF.LoadFile(const FileName: String;
  FilterIndex: Integer);
var
  r: Boolean;
begin
  Screen.Cursor := crHourGlass;
  try
    ShowText('Loading file...');
    RVPP.RVPrint := nil;
    FDocument.Reset;
    case FilterIndex of
      1:
        r := FDocument.MainDoc.RichView.LoadRTF(FileName);
      2:
        r := FDocument.MainDoc.RichView.LoadRVF(FileName);
      3:
        r := FDocument.MainDoc.RichView.LoadDocX(FileName);
      else
        r := False;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
  if not r then
    ShowError('Error opening file');
  ShowText('Preparing to display...');
  Screen.Cursor := crHourGlass;
  try
    AssignVPProperties;
    FDocument.AssignToRVPrint(RVPrint1);
    RVPrint1.FormatPages(rvdoAll);
    RVPP.PageNo := 1;
    RVPP.RVPrint := RVPrint1;
    UpdatePageNavigator;
    btnExport.Enabled := FDocument.MainDoc.RichView.ItemCount > 0;
    ShowText('');
  finally
    Screen.Cursor := crDefault;
  end;
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.btnOpenClick(Sender: TObject);
begin
  if OpenDialog1.Execute then
    LoadFile(OpenDialog1.FileName, OpenDialog1.FilterIndex);
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.DoItemAction(Sender: TCustomRichView;
  ItemAction: TRVItemAction; Item: TCustomRVItemInfo;
  var Text: TRVUnicodeString; RVData: TCustomRVData);
begin
  if (ItemAction = rviaInserting) and (Item is TRVTableItemInfo) then
    TRVTableItemInfo(Item).Options := TRVTableItemInfo(Item).Options + [rvtoCellBelowBorders];
end;

procedure TfrmCustomMakePDF.DoProgress(Sender: TCustomRichView;
  Operation: TRVLongOperation; Stage: TRVProgressStage; PercentDone: Byte);
begin
  if Operation in [rvloRTFRead, rvloRTFWrite] then
    case Stage of
      rvpstgRunning:
        ShowText('Loading file (' + IntToStr(PercentDone) + '%)...');
      rvpstgEnding:
        ShowText('');
    end;
end;
{==================================== Zooming  ================================}
procedure TfrmCustomMakePDF.UpdateZoom;
var
  S:    String;
  Zoom: Integer;
begin
  S := Trim(cmbZoom.Text);
  if S = 'Page width' then
    RVPP.ZoomMode := rvzmPageWidth
  else if S = 'Full page' then
    RVPP.ZoomMode := rvzmFullPage
  else if (S <> '') and (S[Length(S)] = '%') then
  begin
    S := Copy(S, 1, Length(s) - 1);
    Zoom := StrToIntDef(S, 0);
    if Zoom < 10 then
      Zoom := 10
    else if Zoom > 500 then
      Zoom := 500;
    RVPP.SetZoom(zoom);
  end
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.RVPPZoomChanged(Sender: TObject);
begin
  cmbZoom.Text := IntToStr(RVPP.ZoomPercent) + '%';
end;
procedure TfrmCustomMakePDF.cmbZoomKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then
  begin
    UpdateZoom;
    Key := #0;
  end;
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.cmbZoomClick(Sender: TObject);
begin
  UpdateZoom;
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.cmbZoomExit(Sender: TObject);
begin
  UpdateZoom;
end;
{=============================== Page navigator ===============================}
procedure TfrmCustomMakePDF.btnPageNavigateClick(Sender: TObject);
begin
  case (Sender as TSpeedButton).Tag of
    1: RVPP.First;
    2: RVPP.Prev;
    3: RVPP.Next;
    4: RVPP.Last;
  end;
  UpdatePageNavigator;
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.txtPageKeyPress(Sender: TObject; var Key: Char);
begin
  if (Key = #13) then
  begin
    Key := #0;
    GoToPage;
  end;
  if not((Key >= '0') and (Key <= '9')) and (Key >= #32) then
  begin
    Beep;
    Key := #0;
  end;
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.txtPageExit(Sender: TObject);
begin
  GoToPage
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.UpdatePageNavigator;
begin
  btnFirst.Enabled := RVPP.PageNo > 1;
  btnPrior.Enabled := RVPP.PageNo > 1;
  btnNext.Enabled := RVPP.PageNo < RVPP.RVPrint.PagesCount;
  btnLast.Enabled := RVPP.PageNo < RVPP.RVPrint.PagesCount;
  lblPageCount.Caption := 'of ' + IntToStr(RVPP.RVPrint.PagesCount);
  txtPage.Text := IntToStr(RVPP.PageNo);
end;
{------------------------------------------------------------------------------}
procedure TfrmCustomMakePDF.GoToPage;
var
  PageNo: Integer;
begin
  PageNo := StrToIntDef(txtPage.Text, RVPP.PageNo);
  if PageNo < 1 then
    PageNo := 1
  else if PageNo > RVPP.RVPrint.PagesCount then
    PageNo := RVPP.RVPrint.PagesCount;
  RVPP.PageNo := PageNo;
  UpdatePageNavigator;
end;



end.

